MAJOR_VERSION = 1
MINOR_VERSION = 2
PATCH_VERSION = 28
version = '{}.{}.{}'.format(MAJOR_VERSION, MINOR_VERSION, PATCH_VERSION)
__all__  = ['MAJOR_VERSION', 'MINOR_VERSION', 'PATCH_VERSION', 'version']
